from django.forms import ModelForm
from django_filters.rest_framework import FilterSet


def create_model_filter_form(
    model,
    debug=False,
    form_field_class_names=(
        "BooleanField",
        "CharField",
        "DateField",
        "DateTimeField",
        "DecimalField",
        "TextField",
    ),
):

    model_name = model.__name__
    if debug:
        print("model_name:", model_name)

    visible_fields = model._meta.get_fields(include_hidden=False)
    if debug:
        print("visible_fields:", visible_fields)

    form_field_names = [
        f.name
        for f in visible_fields
        if f.editable and f.__class__.__name__ in form_field_class_names
    ]
    if debug:
        print("form_field_names:", form_field_names)

    form_name = model_name + "Form"
    if debug:
        print("form_name:", form_name)

    form_defs = {"Meta": type("Meta", (), {"fields": form_field_names, "model": model})}

    form = type(form_name, (ModelForm,), form_defs)
    if debug:
        print("form:", form)

    return form


def create_model_filterset_class(
    model=None, form=None, debug=False, valid_lookups=None
):

    if form is None and model is None:
        raise Exception("you must pass in a form or model")

    if not form and model:
        form = create_model_filter_form(model)

    model = form.Meta.model
    if debug:
        print("model:", model)

    model_name = model.__name__
    if debug:
        print("model_name:", model_name)

    filter_name = model_name + "Filter"
    if debug:
        print("filter_name:", filter_name)

    visible_fields = model._meta.get_fields(include_hidden=False)
    if debug:
        print("visible_fields:", visible_fields)

    fields = {}
    for field in model._meta.get_fields(include_hidden=False):
        lookups = list(field.__class__.get_lookups().keys())

        if debug:
            print("lookups:", lookups)
        if "date" in lookups:
            for extra_lookup in [
                "date__gt",
                "date__gte",
                "date__lt",
                "date__lte",
                "date__range",
            ]:
                if extra_lookup not in lookups:
                    lookups.append(extra_lookup)

        if valid_lookups:
            lookups = [lookup for lookup in lookups if lookup in valid_lookups]

        if debug:
            print("filtered lookups:", lookups)
        if lookups:
            fields[field.name] = lookups

    if debug:
        print("fields:", fields)

    filter_defs = {
        "Meta": type("Meta", (), {"fields": fields, "model": model, "form": form})
    }
    if debug:
        print("filter_defs:", filter_defs)

    filterset_class = type(filter_name, (FilterSet,), filter_defs)
    if debug:
        print("filterset_class:", filterset_class)

    return filterset_class



from collections import OrderedDict


def get_field_keys(fields, path=""):
    previous = path + "." if path else ""
    results = []

    if hasattr(fields, "_meta"):
        fields = OrderedDict(
            [(field.name, field) for field in fields._meta.get_fields()]
        )

    for field_name, field in fields.items():
        if (
            field.__class__.__name__ == "NestedSerializer"
            or field.__class__.__name__ == "OrderedDict"
            or field.__class__.__name__ == "dict"
        ):
            subobj = field.fields if hasattr(field, "fields") else field
            for result in get_field_keys(subobj, previous + field_name):
                results.append(result)
        else:
            results.append(previous + field_name)
    return results



from collections import Counter
from collections import OrderedDict
from rest_framework_csv.renderers import CSVRenderer
from drf_renderer_xlsx.renderers import XLSXRenderer
import simple_env as se
from .get_field_keys import get_field_keys

DEBUG = se.get("DEBUG_DRF_MORE") or False


class NonPaginatedCSVRenderer(CSVRenderer):
    format = "csv (non-paginated)"

    def render(self, data, accepted_media_type, renderer_context):
        response = renderer_context["response"]

        if response.status_code != 200:
            return ""

        view = renderer_context["view"]
        request = renderer_context["request"]

        queryset = view.queryset or view.get_queryset()
        query_params = request.query_params.dict()
        query_params.pop("format", None)

        # ignore the page parameter
        query_params.pop("page", None)

        # ignore the fields parameter
        query_params.pop("fields", None)

        # filter out keys with empty strings
        filters = {k: v for k, v in query_params.items() if v != ""}

        queryset = queryset.filter(**filters)

        serializer = view.serializer_class

        data = serializer(queryset, context={"request": request}, many=True).data

        return super(NonPaginatedCSVRenderer, self).render(
            data, accepted_media_type, renderer_context
        )


def get(data, path):
    try:
        original_data = data
        for part in path.split("."):
            if data is None or part not in data:
                return None
            data = data[part]
            if data is None:
                return None
        return data
    except Exception as e:
        print("original_data:", original_data)
        print("path:", path)
        raise e


class NonPaginatedXLSXRenderer(XLSXRenderer):
    format = "xlsx (non-paginated)"

    def render(self, data, accepted_media_type, renderer_context):
        response = renderer_context["response"]

        if response.status_code != 200:
            return ""

        view = renderer_context["view"]
        request = renderer_context["request"]

        queryset = view.queryset or view.get_queryset()
        query_params = request.query_params.dict()
        query_params.pop("format", None)

        # ignore the page parameter
        query_params.pop("page", None)

        # ignore the fields parameter
        fields = query_params.pop("fields", None)

        # filter out keys with empty strings
        filters = {k: v for k, v in query_params.items() if v != ""}

        queryset = queryset.filter(**filters)

        serializer = view.serializer_class

        data = serializer(queryset, context={"request": request}, many=True).data
        if DEBUG:
            print("pre-cleaning len(data)", len(data))

        keys = []
        for row in data:
            previous = None
            for key in get_field_keys(row):
                if key not in keys:
                    if previous:
                        index = keys.index(previous) + 1
                        keys = keys[:index] + [key] + keys[index:]
                    else:
                        keys.append(key)
                previous = key

        # remove any null foreign keys
        keys = [
            key
            for key in keys
            if not any([other_key.startswith(key + ".") for other_key in keys])
        ]

        if DEBUG:
            print("keys:", keys)
            print("len(keys):", len(keys))

        # reformat the input data
        flat_data = []
        for row in data:
            flat_row = []
            for key in keys:
                value = get(row, key)
                # don't want empty dicts because that throws off XLSXRenderer
                if isinstance(value, dict) and len(value) == 0:
                    value = None
                flat_row.append((key, value))
            flat_data.append(OrderedDict(flat_row))

        if DEBUG:
            print("post-flattening flat_data[0]", flat_data[0])
            print("post-flattening flat_data[1]", flat_data[1])
            print("post-flattening len(flat_data)", len(flat_data))

        if fields:
            fields = fields.split(",")

            # sort column titles to match how
            # XLSXRenderer will return the data
            column_titles = sorted(fields, key=lambda field: keys.index(field))
        else:
            column_titles = keys

        if DEBUG:
            print("column_titles:", column_titles)

        if hasattr(view, "column_header"):
            if "titles" not in view.column_header:
                view.column_header["titles"] = column_titles
        else:
            view.column_header = {"titles": column_titles}

        return super(NonPaginatedXLSXRenderer, self).render(
            flat_data, accepted_media_type, renderer_context
        )



from setuptools import setup
from os import path

this_directory = path.abspath(path.dirname(__file__))

with open(path.join(this_directory, "README.md")) as f:
    long_description = f.read()

setup(
    name="djangorestframework-more",
    packages=["rest_framework_more"],
    package_dir={"rest_framework_more": "rest_framework_more"},
    package_data={
        "rest_framework_more": [
            "__init__.py",
            "filters.py",
            "mixins.py",
            "renderers.py",
            "serializers.py",
            "viewsets.py",
        ]
    },
    version="0.15.0",
    description="More Cool Django Rest Framework Stuff",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Daniel J. Dufour",
    author_email="daniel.j.dufour@gmail.com",
    url="https://github.com/DanielJDufour/djangorestframework-more",
    download_url="https://github.com/DanielJDufour/djangorestframework-more/tarball/download",
    keywords=[
        "csv",
        "django",
        "django-rest-framework",
        "drf",
        "djangorestframework-csv",
        "excel",
        "pagination",
        "xlsx",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "djangorestframework-csv",
        "drf-renderer-xlsx",
        "prune",
        "simple-env",
    ],
)



from rest_framework.response import Response
from prune import prune
from .get_field_keys import get_field_keys
from .subpaths import subpaths


class FileNameMixin(object):
    def get_filename(self, request, response):
        try:
            return request.path_info.replace("/", "_").strip("_")
        except:
            pass

        try:
            return self.queryset.model.__name__
        except:
            return "download"

    def finalize_response(self, request, response, *args, **kwargs):
        response = super(FileNameMixin, self).finalize_response(
            request, response, *args, **kwargs
        )
        if isinstance(response, Response):
            if "xlsx" in response.accepted_renderer.format:
                filename = self.get_filename(request, response) + ".xlsx"
            elif "csv" in response.accepted_renderer.format:
                filename = self.get_filename(request, response) + ".csv"
            else:
                filename = None

            if filename:
                response["content-disposition"] = f"attachment; filename={filename}"

        return response


class FieldsMixin(object):
    def to_representation(self, *args, **kwargs):
        ret = super(FieldsMixin, self).to_representation(*args, **kwargs)

        try:
            fields = self.context["request"].query_params.get("fields")
            if fields:
                fields_to_keep = fields.split(",")

                all_fields = get_field_keys(self.fields, "")

                remove_these_fields = []
                for field in all_fields:
                    for subpath in subpaths(field):
                        if subpath in remove_these_fields:
                            break

                        if not any(
                            [
                                f == subpath or f.startswith(subpath + ".")
                                for f in fields_to_keep
                            ]
                        ):
                            remove_these_fields.append(subpath)
                            break

                prune(ret, remove_these_fields)
        except Exception as e:
            print(e)

        return ret



import re


def subpaths(path):
    results = []
    indexes = [m.start() for m in re.finditer("\.", path)]
    for index in indexes + [len(path)]:
        results.append(path[0:index])
    return results
